import os
import glob
from typing import List, Dict
import importlib.machinery
import importlib.util
from pymergen.plugin.registry import PluginRegistry
from pymergen.plugin.plugin import Plugin


class PluginManager:

    CATEGORY_COLLECTOR = "collector"

    def __init__(self, context):
        self._context = context
        self._paths = None
        self._init_registry()

    def _init_registry(self):
        self._registry = PluginRegistry()
        self._registry.add_category(self.CATEGORY_COLLECTOR)

    @property
    def paths(self) -> List:
        if self._paths is None:
            self._paths = list()
            self._paths.append(os.path.dirname(__file__))
            if self._context.plugin_path is not None:
                self._paths.append(self._context.plugin_path)
        return self._paths

    @property
    def registry(self) -> PluginRegistry:
        return self._registry

    def load(self):
        for path in self.paths:
            for category in self._registry.categories:
                for plugin_file in glob.glob(os.path.join(path, category, '*', 'plugin.py')):
                    loader = importlib.machinery.SourceFileLoader(category, plugin_file)
                    plugin_spec = importlib.util.spec_from_loader(loader.name, loader)
                    plugin_module = importlib.util.module_from_spec(plugin_spec)
                    loader.exec_module(plugin_module)
                    plugin = plugin_module.Plugin()
                    self._registry.add_plugin(category, plugin.engine, plugin)

    def get_collector_plugin(self, engine: str) -> Plugin:
        return self._registry.get_plugin(self.CATEGORY_COLLECTOR, engine)

    def get_collector_plugins(self) -> Dict[str, Plugin]:
        return self._registry.get_plugins(self.CATEGORY_COLLECTOR)



from typing import List
from pymergen.entity.entity import Entity
from pymergen.entity.suite import EntitySuite
from pymergen.controller.group import ControllerGroup
from pymergen.collector.collector import Collector


class EntityPlan(Entity):

    def __init__(self):
        super().__init__()
        self._suites = list()
        self._cgroups = list()
        self._collectors = list()

    @property
    def suites(self) -> List[EntitySuite]:
        return self._suites

    @suites.setter
    def suites(self, values: List[EntitySuite]) -> None:
        self._suites = list()
        for value in values:
            self.add_suite(value)

    def add_suite(self, value: EntitySuite) -> None:
        value.parent = self
        self._suites.append(value)

    @property
    def cgroups(self) -> List[ControllerGroup]:
        return self._cgroups

    @cgroups.setter
    def cgroups(self, values: List[ControllerGroup]) -> None:
        self._cgroups = values

    @property
    def collectors(self) -> List[Collector]:
        return self._collectors

    @collectors.setter
    def collectors(self, values: List[Collector]) -> None:
        self._collectors = values

    def add_collector(self, value: Collector) -> None:
        self._collectors.append(value)

    def dir_name(self) -> str:
        return "plan_{plan}".format(plan=self.name)

    def short_name(self) -> str:
        return "Plan[{plan}]".format(plan=self.name)

    def long_name(self) -> str:
        return self.short_name()



import os
import glob
import json
from pathlib import Path
from typing import List, Dict
from collections import defaultdict
from pymergen.entity.plan import EntityPlan
from pymergen.core.context import Context
from pymergen.core.executor import ControllingExecutor
from pymergen.core.executor import CollectingExecutor
from pymergen.core.executor import ReplicatingExecutor
from pymergen.core.executor import ConcurrentExecutor
from pymergen.core.executor import IteratingExecutor
from pymergen.core.executor import ParallelExecutor
from pymergen.core.executor import ProcessExecutor


class Runner:

    REPORT_FILES = "files"

    def __init__(self, context: Context):
        self._context = context

    @property
    def context(self) -> Context:
        return self._context

    # Executor hierarchy:
    # Plan {  Controller > Replication }
    #   Suite { Replication > [Collection] > Concurrency }
    #       Case { Replication > Iteration > [Collection] > Parallelism }
    #           Command
    def run(self, plans: List[EntityPlan]) -> None:
        for plan in plans:
            plan_re = ReplicatingExecutor(self.context, plan)
            plan_cne = ControllingExecutor(self.context, plan, plan.cgroups)
            plan_cne.add_child(plan_re)
            for suite in plan.suites:
                suite_cce = ConcurrentExecutor(self.context, suite)
                suite_re = ReplicatingExecutor(self.context, suite)
                plan_re.add_child(suite_re)
                # If a suite is configured with concurrency, then we need to encapsulate all child cases for collection.
                if suite.config.concurrency is True:
                    suite_cle = CollectingExecutor(self.context, suite, plan.collectors, plan.cgroups)
                    suite_cle.add_child(suite_cce)
                    suite_re.add_child(suite_cle)
                else:
                    suite_re.add_child(suite_cce)
                for case in suite.cases:
                    case_pe = ParallelExecutor(self.context, case)
                    for command in case.commands:
                        command_pe = ProcessExecutor(self.context, command)
                        case_pe.add_child(command_pe)
                    case_ie = IteratingExecutor(self.context, case)
                    # No suite concurrency configured means that we can run collectors for each child case.
                    if suite.config.concurrency is False:
                        case_cle = CollectingExecutor(self.context, case, plan.collectors, plan.cgroups)
                        case_cle.add_child(case_pe)
                        case_ie.add_child(case_cle)
                    else:
                        case_ie.add_child(case_pe)
                    case_re = ReplicatingExecutor(self.context, case)
                    case_re.add_child(case_ie)
                    suite_cce.add_child(case_re)
            plan_cne.execute(None)

    def report(self, options: Dict) -> None:
        report = dict()
        if options.get(self.REPORT_FILES):
            report[self.REPORT_FILES] = self._report_files()
        if len(report) > 0:
            print(json.dumps(report, indent=4))

    def _report_files(self) -> Dict:
        report = defaultdict(lambda: defaultdict(list))
        run_path = os.path.abspath(self.context.run_path)
        files = glob.glob("{run_path}/**/*".format(run_path=run_path), recursive=True)
        for file in files:
            if os.path.isfile(file):
                file_name = Path(file).stem
                file_name_parts = file_name.split(".")
                file_name_category = file_name_parts[0]
                report[file_name_category][file_name].append(file)
        return report



import re
from typing import Self, List
from pymergen.entity.config import EntityConfig


class Entity:

    def __init__(self):
        self._name = None
        self._config = EntityConfig()
        self._parent = None
        self._pre = list()
        self._post = list()

    @property
    def config(self) -> EntityConfig:
        return self._config

    @property
    def parent(self) -> Self:
        return self._parent

    @parent.setter
    def parent(self, value: Self) -> None:
        self._parent = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        if re.match('^[a-z0-9_\\-]+$', value, re.IGNORECASE) is None:
            raise Exception("Name {v} can only contain alpha-numeric characters, underscores, or dashes.".format(v=value))
        self._name = value

    @property
    def pre(self) -> List[Self]:
        return self._pre

    @pre.setter
    def pre(self, values: List[Self]) -> None:
        self._pre = list()
        for value in values:
            self.add_pre(value)

    def add_pre(self, value: Self) -> None:
        value.parent = self
        self._pre.append(value)

    @property
    def post(self) -> List[Self]:
        return self._post

    @post.setter
    def post(self, values: List[Self]) -> None:
        self._post = list()
        for value in values:
            self.add_post(value)

    def add_post(self, value: Self) -> None:
        value.parent = self
        self._post.append(value)

    def __str__(self) -> str:
        return self.short_name()

    def short_name(self) -> str:
        raise NotImplementedError()

    def long_name(self) -> str:
        raise NotImplementedError()



import os
import time
import json


class StatTimer:

    def __init__(self):
        self._active = False
        self._started_at = None
        self._stopped_at = None
        self._duration = None

    @property
    def started_at(self) -> float:
        return self._started_at

    @property
    def stopped_at(self) -> float:
        return self._stopped_at

    @property
    def duration(self) -> float:
        if self._duration is None:
            self._duration = self._stopped_at - self._started_at
            self._duration = round(self._duration, 2)
        return self._duration

    def start(self) -> None:
        if self._active:
            raise Exception("Timer is already active")
        self._active = True
        self._started_at = time.time()

    def stop(self) -> None:
        if not self._active:
            raise Exception("Timer is not active")
        self._active = False
        self._stopped_at = time.time()

    def log(self, path: str) -> None:
        data = {
            "started_at": self.started_at,
            "stopped_at": self.stopped_at,
            "duration": self.duration,
        }
        log_file_path = os.path.join(path, "stat.timer.json")
        with open(log_file_path, "w") as fh:
            fh.write("{data}\n".format(data=json.dumps(data)))
            fh.flush()


class Stat:

    def __init__(self):
        self._timer = StatTimer()

    @property
    def timer(self) -> StatTimer:
        return self._timer

    def start(self) -> None:
        self._timer.start()

    def stop(self) -> None:
        self._timer.stop()

    def log(self, path: str) -> None:
        self._timer.log(path)



from typing import Dict
from pymergen.core.context import Context


class Collector:

    def __init__(self):
        self._name = None
        self._context = None

    def parse(self, config: Dict) -> None:
        self._name = config.get("name")

    @property
    def name(self) -> int:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def context(self) -> Context:
        return self._context

    @context.setter
    def context(self, value: Context) -> None:
        self._context = value

    def configure(self, config: Dict) -> None:
        pass

    def start(self, value: Context) -> None:
        raise NotImplementedError()

    def stop(self) -> None:
        raise NotImplementedError()



import os
import yaml
from typing import Dict, Any


class Plugin:

    def __init__(self):
        self._engine = None

    @property
    def engine(self) -> str:
        raise NotImplementedError()

    def schema(self, version: str) -> Dict:
        raise NotImplementedError()

    def implementation(self, config: Dict) -> Any:
        raise NotImplementedError()

    def _engine(self, path: str) -> str:
        if self._engine is None:
            self._engine = os.path.basename(os.path.dirname(path))
        return self._engine

    def _schema(self, path: str, version: str) -> Dict:
        schema_path = os.path.join(os.path.dirname(path), "schema_{version}.yaml".format(version=version))
        if not os.path.exists(schema_path):
            raise Exception("Plugin schema path {schema_path} not found".format(schema_path=schema_path))
        with open(schema_path, "r") as f:
            schema = yaml.safe_load(f)
        return schema



from typing import List
from pymergen.entity.command import EntityCommand
from pymergen.controller.controller import Controller


class ControllerGroup:

    DIR_BASE = "/sys/fs/cgroup"

    def __init__(self, name: str):
        self._name = name
        self._become_cmd = None
        self._controllers = list()

    @property
    def name(self) -> str:
        return self._name

    @property
    def become_cmd(self) -> str:
        return self._become_cmd

    @become_cmd.setter
    def become_cmd(self, value: str) -> None:
        self._become_cmd = value

    @property
    def controllers(self) -> List[Controller]:
        return self._controllers

    @controllers.setter
    def controllers(self, values: List[Controller]) -> None:
        self._controllers = values

    def add_controller(self, value: Controller) -> None:
        self._controllers.append(value)

    def builders(self) -> List[EntityCommand]:
        commands = list()
        c = EntityCommand()
        c.name = "cgcreate_{cgroup}".format(cgroup=self.name)
        c.cmd = "cgcreate -g {controllers}:{cgroup}".format(
            controllers=",".join(self._controller_names()),
            cgroup=self.name
        )
        c.become_cmd = self.become_cmd
        commands.append(c)
        for controller in self.controllers:
            for key, value in controller.limits.items():
                c = EntityCommand()
                c.name = "cgset_{cgroup}_{controller}_{key}".format(cgroup=self.name, controller=controller.name, key=key)
                c.cmd = "cgset -r {controller}.{key}={value} {cgroup}".format(
                    controller=controller.name,
                    key=key, value=value,
                    cgroup=self.name)
                c.become_cmd = self.become_cmd
                commands.append(c)
        return commands

    def destroyers(self) -> List[EntityCommand]:
        commands = list()
        c = EntityCommand()
        c.name = "cgdelete_{cgroup}".format(cgroup=self.name)
        c.cmd = "cgdelete -g {controllers}:{cgroup}".format(
            controllers=",".join(self._controller_names()),
            cgroup=self.name
        )
        c.become_cmd = self.become_cmd
        commands.append(c)
        return commands

    def _controller_names(self) -> List[str]:
        return [c.name for c in self.controllers]



from typing import Dict


class EntityConfig:

    ITERATION_TYPE_PRODUCT = "product"
    ITERATION_TYPE_ZIP = "zip"

    def __init__(self):
        self._replication: int = 1
        self._concurrency: bool = False
        self._parallelism: int = 1
        self._iteration: str = self.ITERATION_TYPE_PRODUCT
        self._params: dict = dict()
        self._iters: dict = dict()

    @property
    def replication(self) -> int:
        return self._replication

    @replication.setter
    def replication(self, value: int) -> None:
        self._replication = value

    @property
    def concurrency(self) -> bool:
        return self._concurrency

    @concurrency.setter
    def concurrency(self, value: bool) -> None:
        self._concurrency = value

    @property
    def parallelism(self) -> int:
        return self._parallelism

    @parallelism.setter
    def parallelism(self, value: int) -> None:
        self._parallelism = value

    @property
    def iteration(self) -> str:
        return self._iteration

    @iteration.setter
    def iteration(self, value: str) -> None:
        self._iteration = value

    @property
    def params(self) -> Dict:
        return self._params

    @params.setter
    def params(self, values: Dict) -> None:
        self._params = values

    @property
    def iters(self) -> Dict:
        return self._iters

    @iters.setter
    def iters(self, values: Dict) -> None:
        self._iters = values



from typing import List
from pymergen.entity.entity import Entity
from pymergen.entity.case import EntityCase


class EntitySuite(Entity):

    def __init__(self):
        super().__init__()
        self._cases = list()

    @property
    def cases(self) -> List[EntityCase]:
        return self._cases

    @cases.setter
    def cases(self, values: List[EntityCase]) -> None:
        self._cases = list()
        for value in values:
            self.add_case(value)

    def add_case(self, value: EntityCase) -> None:
        value.parent = self
        self._cases.append(value)

    def dir_name(self) -> str:
        return "suite_{suite}".format(suite=self.name)

    def short_name(self) -> str:
        return "Suite[{suite}]".format(suite=self.name)

    def long_name(self) -> str:
        return "Plan[{plan}] > Suite[{suite}]".format(plan=self.parent.name, suite=self.name)



from pymergen.controller.controller import Controller
from pymergen.controller.controller import ControllerCpuset
from pymergen.controller.controller import ControllerCpu
from pymergen.controller.controller import ControllerIo
from pymergen.controller.controller import ControllerMemory
from pymergen.controller.controller import ControllerHugeTlb
from pymergen.controller.controller import ControllerPids
from pymergen.controller.controller import ControllerRdma
from pymergen.controller.controller import ControllerMisc


class ControllerFactory:

    @staticmethod
    def instance(name) -> Controller:
        if name == Controller.TYPE_CPUSET:
            return ControllerCpuset()
        if name == Controller.TYPE_CPU:
            return ControllerCpu()
        if name == Controller.TYPE_IO:
            return ControllerIo()
        if name == Controller.TYPE_MEMORY:
            return ControllerMemory()
        if name == Controller.TYPE_HUGETLB:
            return ControllerHugeTlb()
        if name == Controller.TYPE_PIDS:
            return ControllerPids()
        if name == Controller.TYPE_RDMA:
            return ControllerRdma()
        if name == Controller.TYPE_MISC:
            return ControllerMisc()
        raise Exception("Controller name {name} is not recognized".format(name=name))



from typing import Dict
from pymergen.collector.collector import Collector
from pymergen.entity.command import EntityCommand
from pymergen.core.executor import ExecutorContext, AsyncProcessExecutor


class CollectorProcess(Collector):

    def __init__(self):
        super().__init__()
        self._executor = None
        self._cmd = None
        self._become_cmd = None
        self._shell = False
        self._shell_executable = None
        self._pipe_stdout = None
        self._pipe_stderr = None

    def parse(self, config: Dict) -> None:
        super().parse(config)
        self._cmd = config.get("cmd", None)
        self._become_cmd = config.get("become_cmd", None)
        self._shell = config.get("shell", False)
        self._shell_executable = config.get("shell_executable", None)
        self._pipe_stdout = config.get("pipe_stdout", None)
        self._pipe_stderr = config.get("pipe_stderr", None)

    @property
    def cmd(self) -> str:
        return self._cmd

    @cmd.setter
    def cmd(self, value: str) -> None:
        self._cmd = value

    @property
    def become_cmd(self) -> str:
        return self._become_cmd

    @become_cmd.setter
    def become_cmd(self, value: str) -> None:
        self._become_cmd = value
        
    @property
    def shell(self) -> bool:
        return self._shell

    @shell.setter
    def shell(self, value: bool) -> None:
        self._shell = value

    @property
    def shell_executable(self) -> str:
        return self._shell_executable

    @shell_executable.setter
    def shell_executable(self, value: str) -> None:
        self._shell_executable = value

    @property
    def pipe_stdout(self) -> str:
        return self._pipe_stdout

    @pipe_stdout.setter
    def pipe_stdout(self, value: str) -> None:
        self._pipe_stdout = value

    @property
    def pipe_stderr(self) -> str:
        return self._pipe_stderr

    @pipe_stderr.setter
    def pipe_stderr(self, value: str) -> None:
        self._pipe_stderr = value

    def command(self) -> EntityCommand:
        command = EntityCommand()
        command.name = self.name
        command.cmd = self.cmd
        command.become_cmd = self.become_cmd
        command.shell = self.shell
        command.shell_executable = self.shell_executable
        command.pipe_stdout = self.pipe_stdout
        command.pipe_stderr = self.pipe_stderr
        return command

    def start(self, parent_context: ExecutorContext) -> None:
        self._executor = AsyncProcessExecutor(self.context, self.command())
        self._executor.execute(parent_context)

    def stop(self) -> None:
        self._executor.execute_stop()



from typing import List, Dict, Union


class Controller:

    TYPE_CPUSET = "cpuset"
    TYPE_CPU = "cpu"
    TYPE_IO = "io"
    TYPE_MEMORY = "memory"
    TYPE_HUGETLB = "hugetlb"
    TYPE_PIDS = "pids"
    TYPE_RDMA = "rdma"
    TYPE_MISC = "misc"

    def __init__(self, name):
        self._name = name
        self._limits = dict()
        self._stat_files = list()

    @property
    def name(self) -> str:
        return self._name

    @property
    def limits(self) -> Dict:
        return self._limits

    @property
    def stat_files(self) -> List:
        return self._stat_files

    @stat_files.setter
    def stat_files(self, values: List) -> None:
        self._stat_files = values

    def add_stat_file(self, value: str) -> None:
        self._stat_files.append(value)

    @limits.setter
    def limits(self, values: Dict) -> None:
        self._limits = values

    def add_limit(self, key: str, value: Union[int, float, str]) -> None:
        self._limits[key] = value


class ControllerCpuset(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_CPUSET)


class ControllerCpu(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_CPU)
        self.add_stat_file("cpu.stat")


class ControllerIo(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_IO)
        self.add_stat_file("io.stat")


class ControllerMemory(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_MEMORY)
        self.add_stat_file("memory.stat")
        self.add_stat_file("memory.numa_stat")


class ControllerHugeTlb(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_HUGETLB)
        self.add_stat_file("hugetlb.1GB.numa_stat")
        self.add_stat_file("hugetlb.2MB.numa_stat")


class ControllerPids(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_PIDS)


class ControllerRdma(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_RDMA)


class ControllerMisc(Controller):

    def __init__(self):
        super().__init__(Controller.TYPE_MISC)




from typing import Dict
from pymergen.plugin.plugin import Plugin as BasePlugin
from pymergen.collector.process import CollectorProcess


class Plugin(BasePlugin):

    @property
    def engine(self) -> str:
        return super()._engine(__file__)

    def schema(self, version: str) -> Dict:
        return super()._schema(__file__, version)

    def implementation(self, config: Dict) -> CollectorProcess:
        collector = CollectorProcess()
        collector.parse(config)
        return collector



import os
import logging


class Logger:

    _logger = None

    @classmethod
    def logger(cls, context) -> logging.Logger:

        if cls._logger is not None:
            return cls._logger

        formatter = logging.Formatter('%(asctime)s - %(name)s - %(process)d - %(thread)d - %(levelname)s - %(filename)s - %(message)s')

        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(context.log_level)
        stream_handler.setFormatter(formatter)

        file_handler = logging.FileHandler(os.path.join(context.run_path, "run.runner.log"))
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)

        cls._logger = logging.getLogger("pymergen")
        cls._logger.setLevel(logging.DEBUG)
        cls._logger.addHandler(stream_handler)
        cls._logger.addHandler(file_handler)

        return cls._logger



import threading
from typing import List
from pymergen.core.context import Context


class Thread:

    def __init__(self, context: Context):
        self._context = context
        self._type = None
        self._thread = None

    @property
    def context(self) -> Context:
        return self._context

    def run(self, target, args: List) -> None:
        self._type = type(target).__name__
        self.context.logger.debug("Thread[{type}] Run[name={name}]".format(type=self._type, name=target.name))
        self._thread = threading.Thread(name=target.name, target=target.run, args=args)
        self._thread.start()

    def join(self) -> None:
        self.context.logger.debug("Thread[{type}] Join[name={name}]".format(type=self._type, name=self._thread.name))
        self._thread.join()



import sys
import shutil
import os
import logging
from datetime import datetime
from pymergen.core.logger import Logger
from pymergen.plugin.manager import PluginManager


class Context:

    def __init__(self, args):
        self._plan_path = args.plan_path
        self._work_path = args.work_path
        self._plugin_path = args.plugin_path
        self._run_path = os.path.join(self._work_path, self._generate_run_path())
        self._log_level = args.log_level
        self._filter_plan = args.filter_plan
        self._filter_suite = args.filter_suite
        self._filter_case = args.filter_case
        self._prepare()
        self._init_logger()
        self._plugin_manager = None

    @property
    def plan_path(self) -> str:
        return self._plan_path

    @property
    def work_path(self) -> str:
        return self._work_path

    @property
    def plugin_path(self) -> str:
        return self._plugin_path

    @property
    def run_path(self) -> str:
        return self._run_path

    @property
    def log_level(self) -> str:
        return self._log_level

    @property
    def filter_plan(self) -> str:
        return self._filter_plan

    @property
    def filter_suite(self) -> str:
        return self._filter_suite

    @property
    def filter_case(self) -> str:
        return self._filter_case

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @property
    def plugin_manager(self) -> PluginManager:
        if self._plugin_manager is None:
            self._plugin_manager = PluginManager(self)
            self._plugin_manager.load()
        return self._plugin_manager

    def validate(self):
        if sys.platform != "linux":
            raise Exception("Linux support only")
        for binary in ["cgcreate", "cgset", "cgdelete", "cgexec", "perf"]:
            if shutil.which(binary) is None:
                raise Exception("Command {binary} not found".format(binary=binary))
        if not os.path.exists(self._plan_path):
            raise Exception("Plan path {path} does not exist".format(path=self._plan_path))

    def _prepare(self) -> None:
        if not os.path.isdir(self.work_path):
            os.mkdir(self.work_path)
        if not os.path.isdir(self.run_path):
            os.mkdir(self.run_path)

    def _init_logger(self):
        self._logger = Logger.logger(self)
        for handler in self._logger.handlers:
            if type(handler) is logging.StreamHandler:
                handler.setLevel(self._log_level)

    @staticmethod
    def _generate_run_path() -> str:
        return datetime.strftime(datetime.now(), "%Y%m%d_%H%M%S")



import argparse
from pymergen.core.parser import Parser
from pymergen.core.context import Context
from pymergen.core.runner import Runner

parser = argparse.ArgumentParser()
parser.add_argument("-p", "--plan-path", action="store", type=str, required=True)
parser.add_argument("-w", "--work-path", action="store", type=str, required=True)
parser.add_argument("--plugin-path", action="store", type=str, required=False)
parser.add_argument("--filter-plan", action="store", type=str, required=False, metavar="REGEX", help="Filter plans by name")
parser.add_argument("--filter-suite", action="store", type=str, required=False, metavar="REGEX", help="Filter suites by name")
parser.add_argument("--filter-case", action="store", type=str, required=False, metavar="REGEX", help="Filter cases by name")
parser.add_argument("-l", "--log-level", action="store", type=str.upper, choices=["DEBUG", "INFO", "WARN", "ERROR"], default="INFO")
parser.add_argument("--report-files", action="store_true", default=False)
args = parser.parse_args()

context = Context(args)
context.validate()

parser = Parser(context)
parser.load()
plans = parser.parse()

runner = Runner(context)
runner.run(plans)
runner.report({
    runner.REPORT_FILES: args.report_files
})



import itertools
import subprocess
import shlex
import signal
import threading
import time
from pymergen.core.context import Context
from pymergen.entity.command import EntityCommand


class Process:

    def __init__(self, context: Context):
        self._context = context
        self._command = None
        self._process = None
        self._stdout = None
        self._stderr = None

    @property
    def context(self) -> Context:
        return self._context

    @property
    def command(self) -> EntityCommand:
        return self._command

    @command.setter
    def command(self, command: EntityCommand):
        self._command = command

    def run(self) -> None:
        self.start()
        self.wait()

    def start(self) -> None:
        try:
            self.context.logger.debug("{n} Execute[{cmd}]".format(n=self._command, cmd=self._command.cmd))
            if self._command.pipe_stdout:
                self._stdout = open(self._command.pipe_stdout, "w")
            if self._command.pipe_stderr:
                self._stderr = open(self._command.pipe_stderr, "w")
            self._process = self._popen()
            if self._command.run_time > 0:
                self._timer()
        except Exception as e:
            self.context.logger.error("Failed to start {n} due to {e}".format(n=self._command, e=e))
            if self._command.raise_error:
                raise e

    def signal(self, sig: signal.Signals = signal.SIGINT) -> None:
        if self._process is None:
            self.context.logger.error("No process to signal for {n}".format(n=self._command))
            return
        try:
            self.context.logger.debug("{n} Signal[signal={sig}]".format(n=self._command, sig=sig))
            self._process.send_signal(sig)
        except Exception as e:
            self.context.logger.error("Failed to send signal {sig} to {n} due to {e}".format(sig=sig, n=self._command, e=e))
            if self._process:
                self._process.kill()
            if self._command.raise_error:
                raise e

    def wait(self) -> None:
        if self._process is None:
            self.context.logger.error("No process to wait for {n}".format(n=self._command))
            return
        try:
            stdout, stderr = self._process.communicate(timeout=self._command.timeout)
            if self._command.debug_stdout:
                if self._command.pipe_stdout:
                    self.context.logger.warning("No debugging output will be captured when stdout is piped")
                self.context.logger.debug(stdout)
            if self._command.debug_stderr:
                if self._command.pipe_stderr:
                    self.context.logger.warning("No debugging output will be captured when stderr is piped")
                self.context.logger.debug(stderr)
            self.context.logger.debug("{n} Return[return_code={r}]".format(n=self._command, r=self._process.returncode))
        except subprocess.TimeoutExpired as e:
            self.context.logger.error("Timeout expiration for {n}".format(n=self._command))
            if self._process:
                self._process.kill()
            if self._command.raise_error:
                raise e
        except Exception as e:
            self.context.logger.error("Failed to wait for {n} due to {e}".format(n=self._command, e=e))
            if self._process:
                self._process.kill()
            if self._command.raise_error:
                raise e
        finally:
            if self._stdout:
                self._stdout.close()
            if self._stderr:
                self._stderr.close()

    def _popen(self) -> subprocess.Popen:
        stdout = subprocess.PIPE
        stderr = subprocess.PIPE
        if self._command.shell is True:
            if self._stdout:
                stdout = self._stdout
            if self._stderr:
                stderr = self._stderr
            return subprocess.Popen(self._command.cmd,
                                    shell=True,
                                    executable=self._command.shell_executable,
                                    stdin=None,
                                    stdout=stdout,
                                    stderr=stderr
                                    )
        # shell is False
        # create a list of sub commands by splitting the full command by the pipe character
        cmd_parts = shlex.split(self._command.cmd)
        sub_cmds = list()
        for k, g in itertools.groupby(cmd_parts, lambda x: x == "|"):
            if not k:
                sub_cmds.append(list(g))
        # chain sub commands
        s_curr = None
        s_prev = None
        total_cmds = len(sub_cmds)
        for i, sub_cmd in enumerate(sub_cmds):
            s_curr_stdin = None
            if s_prev is not None:
                s_curr_stdin = s_prev.stdout
            is_last_command = (i == total_cmds - 1)
            if is_last_command and self._stdout:
                stdout = self._stdout
            if is_last_command and self._stderr:
                stderr = self._stderr
            s_curr = subprocess.Popen(sub_cmd,
                                      shell=False,
                                      executable=self._command.shell_executable,
                                      stdin=s_curr_stdin,
                                      stdout=stdout,
                                      stderr=stderr
                                      )
            if s_prev is not None:
                s_prev.stdout.close()
            s_prev = s_curr
        return s_curr

    def _timer(self) -> None:
        self.context.logger.debug(
            "{n} Timer[run_time={run_time}]".format(n=self._command, run_time=self._command.run_time))
        sleep_time = 0
        while sleep_time < self._command.run_time:
            if self._process.poll() is not None:
                self.context.logger.debug("{n} exited with return code {r} before run timer expired".format(n=self._command, r=self._process.returncode))
                return
            time.sleep(1)
            sleep_time += 1
        self.signal()



import os
import re
import glob
import importlib
import cerberus
import pprint
import yaml
from typing import Any, List, Dict
from pymergen.entity.plan import EntityPlan
from pymergen.entity.suite import EntitySuite
from pymergen.entity.case import EntityCase
from pymergen.entity.command import EntityCommand
from pymergen.core.context import Context
from pymergen.controller.factory import ControllerFactory
from pymergen.controller.group import ControllerGroup
from pymergen.collector.collector import Collector


class Parser:

    def __init__(self, context: Context):
        self._context = context
        self._plans = list()
        self._validator = None
        self._init_validator()

    @property
    def context(self) -> Context:
        return self._context

    def load(self):
        path = self.context.plan_path
        if os.path.isdir(path):
            files = glob.glob(os.path.join(path, "*.yaml"))
            if len(files) == 0:
                raise Exception("No YAML files are found in {}".format(path))
            for file in files:
                data = self._load_yaml(file)
                self._validate_document(data, file)
                self._plans.extend(data["plans"])
        if os.path.isfile(path):
            if not path.endswith(".yaml"):
                raise Exception("Only YAML files are supported")
            data = self._load_yaml(path)
            self._validate_document(data, path)
            self._plans.extend(data["plans"])

    def parse(self) -> List[EntityPlan]:
        plans = [self._parse_plan(plan) for plan in self._plans]
        if self.context.filter_plan is not None:
            self.context.logger.debug("Plan filter is {f}".format(f=self.context.filter_plan))
            plans = [plan for plan in plans if re.search(self.context.filter_plan, plan.name, flags=re.IGNORECASE)]
        return plans

    def _parse_plan(self, data: Dict) -> EntityPlan:
        plan = EntityPlan()
        plan.name = data["name"]
        config = data.get("config", {})
        plan.config.replication = config.get("replication", 1)
        plan.config.params = config.get("params", dict())
        plan.config.iters = config.get("iters", dict())
        plan.pre = self._parse_commands(data.get("pre", []))
        plan.post = self._parse_commands(data.get("post", []))
        plan.cgroups = self._parse_cgroups(data.get("cgroups", []))
        plan.collectors = self._parse_collectors(data.get("collectors", []))
        for s in data["suites"]:
            suite = self._parse_suite(s)
            if self.context.filter_suite is not None:
                self.context.logger.debug("Suite filter is {f}".format(f=self.context.filter_suite))
                if re.search(self.context.filter_suite, suite.name, flags=re.IGNORECASE):
                    plan.add_suite(suite)
            else:
                plan.add_suite(suite)
        return plan

    def _parse_suite(self, data: Dict) -> EntitySuite:
        suite = EntitySuite()
        suite.name = data["name"]
        config = data.get("config", {})
        suite.config.replication = config.get("replication", 1)
        suite.config.concurrency = config.get("concurrency", False)
        suite.config.params = config.get("params", dict())
        suite.config.iters = config.get("iters", dict())
        suite.pre = self._parse_commands(data.get("pre", []))
        suite.post = self._parse_commands(data.get("post", []))
        for c in data["cases"]:
            case = self._parse_case(c)
            if self.context.filter_case is not None:
                self.context.logger.debug("Case filter is {f}".format(f=self.context.filter_case))
                if re.search(self.context.filter_case, case.name, flags=re.IGNORECASE):
                    suite.add_case(case)
            else:
                suite.add_case(case)
        return suite

    def _parse_case(self, data: Dict) -> EntityCase:
        case = EntityCase()
        case.name = data["name"]
        config = data.get("config", {})
        case.config.replication = config.get("replication", 1)
        case.config.parallelism = config.get("parallelism", 1)
        case.config.params = config.get("params", dict())
        case.config.iters = config.get("iters", dict())
        case.pre = self._parse_commands(data.get("pre", []))
        case.post = self._parse_commands(data.get("post", []))
        case.commands = self._parse_commands(data.get("commands", []))
        return case

    def _parse_commands(self, data: List[Dict]) -> List[EntityCommand]:
        commands = list()
        for item in data:
            command = EntityCommand()
            command.name = item.get("name")
            command.cmd = item.get("cmd")
            command.become_cmd = item.get("become_cmd", None)
            command.raise_error = item.get("raise_error", True)
            command.run_time = item.get("run_time", 0)
            command.shell = item.get("shell", False)
            command.shell_executable = item.get("shell_executable", None)
            command.timeout = item.get("timeout", None)
            command.pipe_stdout = item.get("pipe_stdout", None)
            command.pipe_stderr = item.get("pipe_stderr", None)
            command.debug_stdout = item.get("debug_stdout", False)
            command.debug_stderr = item.get("debug_stderr", False)
            command.cgroups = item.get("cgroups", list())
            commands.append(command)
        return commands

    def _parse_cgroups(self, data: List) -> List[ControllerGroup]:
        cgroups = list()
        for item in data:
            cgroup = ControllerGroup(item.get("name"))
            cgroup.become_cmd = item.get("become_cmd", None)
            for c in item.get("controllers"):
                controller = ControllerFactory.instance(c.get("name"))
                for limit in c.get("limits", {}):
                    controller.add_limit(limit.get("key"), limit.get("value"))
                cgroup.add_controller(controller)
            cgroups.append(cgroup)
        return cgroups

    def _parse_collectors(self, data: List) -> List[Collector]:
        collectors = list()
        for item in data:
            collector_plugin = self.context.plugin_manager.get_collector_plugin(item.get("engine"))
            collector = collector_plugin.implementation(item)
            collector.context = self.context
            collectors.append(collector)
        return collectors

    def _init_validator(self) -> None:
        try:
            importlib.import_module("cerberus")
            self._validator = cerberus.Validator()
        except ModuleNotFoundError:
            self.context.logger.warning("Unable to load module cerberus")

    def _validate_document(self, document: Dict, path: str) -> None:
        if not self._validator:
            return
        version = document.get("version", "")
        if len(version) == 0:
            raise Exception("Document does not have a version defined")
        schema_file = "schema_{version}.yaml".format(version=version)
        schema_path = os.path.join(os.path.dirname(__file__), "../conf/schema", schema_file)
        if not os.path.exists(schema_path):
            raise Exception("Schema path {schema_path} does not exist".format(schema_path=schema_path))
        schema = self._load_yaml(schema_path)
        # Add collector schemas
        collector_schemas = schema.get('plans').get('schema').get('schema').get('collectors').get('schema').get('anyof')
        collector_schemas.clear()
        for plugin in self.context.plugin_manager.get_collector_plugins().values():
            plugin_schema = plugin.schema(version)
            collector_schemas.append(plugin_schema)
        # Validate document with schema
        self._validator.validate(document, schema)
        if self._validator.errors:
            pprint.pprint(document)
            pprint.pprint(self._validator.errors)
            raise Exception("Failed to validate document {path} due to: {error}".format(
                path=path,
                error=pprint.pformat(self._validator.errors)
            ))

    def _load_yaml(self, path: str) -> Dict:
        with open(path, "r") as f:
            data = yaml.safe_load(f)
        base_dir = os.path.dirname(path)
        data = self._load_includes(data, base_dir)
        return data

    def _load_includes(self, data: Any, base_dir: str) -> Any:
        if isinstance(data, str) and data.startswith("include:") and data.endswith(".yaml"):
            path = self._get_include_path(data, base_dir)
            return self._load_yaml(path)
        if isinstance(data, list):
            return [self._load_includes(item, base_dir) for item in data]
        if isinstance(data, dict):
            return {k: self._load_includes(v, base_dir) for k, v in data.items()}
        return data

    def _get_include_path(self, value: str, base_dir: str) -> str:
        path = re.sub(r"^include:", "", value)
        path = os.path.join(base_dir, path)
        if not os.path.exists(path):
            raise Exception("Failed to load include at {path}".format(path=path))
        return path



from typing import List
from pymergen.entity.entity import Entity
from pymergen.entity.command import EntityCommand


class EntityCase(Entity):

    def __init__(self):
        super().__init__()
        self._commands = list()

    @property
    def commands(self) -> List[EntityCommand]:
        return self._commands

    @commands.setter
    def commands(self, values: List[EntityCommand]) -> None:
        self._commands = list()
        for value in values:
            self.add_command(value)

    def add_command(self, value: EntityCommand) -> None:
        value.parent = self
        self._commands.append(value)

    def dir_name(self) -> str:
        return "case_{case}".format(case=self.name)

    def short_name(self) -> str:
        return "Case[{case}]".format(case=self.name)

    def long_name(self) -> str:
        return "Plan[{plan}] > Suite[{suite}] > Case[{case}]".format(plan=self.parent.parent.name, suite=self.parent.name, case=self.name)



import os
import time
from io import TextIOBase
from typing import Self, List
from datetime import datetime
from pymergen.collector.thread import CollectorThread
from pymergen.core.executor import CollectingExecutorContext


class CollectorControllerGroupFile:

    _instances = dict()

    def __init__(self, path: str, mode: str):
        self._path = path
        self._mode = mode
        self._fh = None

    def __del__(self):
        if self._fh is not None and not self._fh.closed:
            self._fh.close()

    @property
    def fh(self) -> TextIOBase:
        if self._fh is None:
            self._fh = open(self._path, self._mode)
        return self._fh


class CollectorControllerGroupStatParser(CollectorControllerGroupFile):

    @staticmethod
    def instance(path: str, mode: str) -> Self:
        if path not in CollectorControllerGroupStatParser._instances:
            CollectorControllerGroupStatParser._instances[path] = CollectorControllerGroupStatParser(path, mode)
        return CollectorControllerGroupStatParser._instances[path]

    def parse_headers(self) -> List:
        headers = list()
        headers.append("timestamp")
        self.fh.seek(0)
        for i, input_line in enumerate(self.fh.readlines()):
            input_columns = input_line.split(" ")
            input_columns_len = len(input_columns)
            if input_columns_len < 2:
                raise Exception("Unable to parse headers from {path} in line: {input_line}".format(
                    path=self._path,
                    input_line=input_line)
                )
            # First input column is the field prefix, and the rest of the columns contain a part of the field
            # name to on the left side of the equal sign.
            #
            # For example:
            #
            # some avg10=0.00 avg60=0.00 avg300=0.00 total=219731
            # full avg10=0.00 avg60=0.00 avg300=0.00 total=146364
            if "=" in input_columns[1]:
                for j in range(1, input_columns_len):
                    column_parts = input_columns[j].split("=")
                    headers.append("{name}_{subname}".format(
                        name=input_columns[0].strip(),
                        subname=column_parts[0].strip())
                    )
            # Two column format where first column is the field name and the right column is the field value.
            #
            # For example:
            #
            # usage_usec 76128949
            # user_usec 45340836
            # system_usec 30788112
            else:
                headers.append(input_columns[0].strip())
        return headers

    def parse_values(self) -> List:
        values = list()
        values.append(datetime.now().isoformat())
        self.fh.seek(0)
        for i, input_line in enumerate(self.fh.readlines()):
            input_columns = input_line.split(" ")
            input_columns_len = len(input_columns)
            if input_columns_len < 2:
                raise Exception("Unable to parse headers from {path} in line: {input_line}".format(
                    path=self._path,
                    input_line=input_line)
                )
            if "=" in input_columns[1]:
                for j in range(1, input_columns_len):
                    column_parts = input_columns[j].split("=")
                    values.append(column_parts[1].strip())
            else:
                values.append(input_columns[1].strip())
        return values


class CollectorControllerGroupStatLogger(CollectorControllerGroupFile):

    def __init__(self, path: str, mode: str):
        super().__init__(path, mode)
        self._is_first_call = True

    @property
    def is_first_call(self) -> bool:
        result = self._is_first_call
        self._is_first_call = False
        return result

    @staticmethod
    def instance(path: str, mode: str) -> Self:
        if path not in CollectorControllerGroupStatLogger._instances:
            CollectorControllerGroupStatLogger._instances[path] = CollectorControllerGroupStatLogger(path, mode)
        return CollectorControllerGroupStatLogger._instances[path]

    def log_line(self, line: str) -> None:
        self.fh.write("{line}\n".format(line=line))
        self.fh.flush()


class CollectorControllerGroup(CollectorThread):

    def run(self, parent_context: CollectingExecutorContext) -> None:
        time.sleep(self.ramp)
        while self._join is False:
            for cgroup in parent_context.cgroups:
                for controller in cgroup.controllers:
                    for stat_file in controller.stat_files:
                        log_file_path = os.path.join(
                            self._executor.run_path(parent_context),
                            "collector.cgroup_{cgname}_{stat_file}.log".format(
                                cgname=cgroup.name,
                                stat_file=stat_file.replace(".", "_")
                            )
                        )
                        stat_logger = CollectorControllerGroupStatLogger.instance(log_file_path, 'a')
                        stat_file_path = os.path.join(cgroup.DIR_BASE, cgroup.name, stat_file)
                        stat_parser = CollectorControllerGroupStatParser.instance(stat_file_path, 'r')
                        if stat_logger.is_first_call:
                            stat_logger.log_line(" ".join(stat_parser.parse_headers()))
                        stat_logger.log_line(" ".join(stat_parser.parse_values()))
            time.sleep(self.interval)



import copy
import itertools
import os
import re
import threading
from typing import Any, List, Dict, Self
from pymergen.entity.entity import EntityConfig, Entity
from pymergen.entity.command import EntityCommand
from pymergen.entity.case import EntityCase
from pymergen.entity.suite import EntitySuite
from pymergen.entity.plan import EntityPlan
from pymergen.core.context import Context
from pymergen.core.process import Process
from pymergen.core.thread import Thread
from pymergen.core.stat import Stat
from pymergen.controller.group import ControllerGroup
from pymergen.collector.collector import Collector


class ExecutorContext:

    def __init__(self, parent: Self):
        self._parent = parent
        self._entity = None
        self._current = None
        self._prefix = None
        self._exclude_from_path = False

    @property
    def parent(self) -> Self:
        return self._parent

    @property
    def entity(self) -> Entity:
        return self._entity

    @entity.setter
    def entity(self, value: Entity):
        self._entity = value

    @property
    def current(self) -> int:
        return self._current

    @current.setter
    def current(self, value: int) -> None:
        self._current = value

    @property
    def exclude_from_path(self) -> bool:
        return self._exclude_from_path

    def id(self) -> str:
        return "{p}{c:03d}".format(p=self._prefix, c=self._current)


class ControllingExecutorContext(ExecutorContext):

    def __init__(self, parent: Self):
        super().__init__(parent)
        self._prefix = "cne"
        self._exclude_from_path = True


class CollectingExecutorContext(ExecutorContext):

    def __init__(self, parent: Self):
        super().__init__(parent)
        self._prefix = "cle"
        self._exclude_from_path = True
        self._cgroups = None

    @property
    def cgroups(self) -> List[ControllerGroup]:
        return self._cgroups

    @cgroups.setter
    def cgroups(self, values: List[ControllerGroup]) -> None:
        self._cgroups = values


class ReplicatingExecutorContext(ExecutorContext):

    def __init__(self, parent: Self):
        super().__init__(parent)
        self._prefix = "r"


class ConcurrentExecutorContext(ExecutorContext):

    def __init__(self, parent: Self):
        super().__init__(parent)
        self._prefix = "cce"
        self._exclude_from_path = True


class ParallelExecutorContext(ExecutorContext):

    def __init__(self, parent: Self):
        super().__init__(parent)
        self._prefix = "p"


class IteratingExecutorContext(ExecutorContext):

    def __init__(self, parent: Self):
        super().__init__(parent)
        self._prefix = "i"
        self._iters = dict()

    @property
    def iters(self) -> Dict[str, str]:
        return self._iters

    @iters.setter
    def iters(self, value: Dict[str, str]) -> None:
        self._iters = value


class Executor:

    def __init__(self, context: Context, entity: Entity):
        self._context = context
        self._entity = entity
        self._children = list()

    @property
    def context(self) -> Context:
        return self._context

    @property
    def entity(self) -> Entity:
        return self._entity

    @property
    def children(self) -> List:
        return self._children

    @children.setter
    def children(self, values: List) -> None:
        self._children = values

    def add_child(self, value: Self) -> None:
        self._children.append(value)

    def execute(self, parent_context: ExecutorContext) -> None:
        self.execute_main(parent_context)

    def execute_pre(self, parent_context: ExecutorContext) -> None:
        for pre in self.entity.pre:
            pe = ProcessExecutor(self.context, pre)
            pe.execute(parent_context)

    def execute_main(self, parent_context: ExecutorContext) -> None:
        raise NotImplementedError()

    def execute_post(self, parent_context: ExecutorContext) -> None:
        for post in self.entity.post:
            pe = ProcessExecutor(self.context, post)
            pe.execute(parent_context)

    # Do NOT cache this logic. Parallel executor requires things to stay independent if we don't want to deal with cloned objects.
    def run_path(self, parent_context: ExecutorContext) -> str:
        names = dict()
        while parent_context is not None:
            if parent_context.exclude_from_path is True:
                parent_context = parent_context.parent
                continue
            entity = parent_context.entity.name
            if parent_context.entity.name not in names:
                names[entity] = list()
            names[entity].append(parent_context.id())
            parent_context = parent_context.parent
        dirs = list()
        for entity in reversed(names):
            dirs.append(entity)
            for name in reversed(names[entity]):
                dirs.append(name)
        run_path = os.path.join(self.context.run_path, *dirs)
        os.makedirs(run_path, exist_ok=True)
        return run_path

    def stat(self):
        return Stat()


class ControllingExecutor(Executor):

    def __init__(self, context: Context, entity: Entity, cgroups: List[ControllerGroup]):
        super().__init__(context, entity)
        self._cgroups = cgroups

    @property
    def cgroups(self) -> List[ControllerGroup]:
        return self._cgroups

    @cgroups.setter
    def cgroups(self, values: List[ControllerGroup]) -> None:
        self._cgroups = values

    def execute_main(self, parent_context: ExecutorContext) -> None:
        try:
            self._build(parent_context)
            for child in self.children:
                context = ControllingExecutorContext(parent_context)
                context.entity = self.entity
                child.execute(context)
        finally:
            self._destroy(parent_context)

    def _build(self, parent_context: ExecutorContext):
        for cgroup in self.cgroups:
            for command in cgroup.builders():
                context = ControllingExecutorContext(parent_context)
                context.entity = self.entity
                pe = ProcessExecutor(self.context, command)
                pe.execute(context)

    def _destroy(self, parent_context: ExecutorContext):
        for cgroup in self.cgroups:
            for command in cgroup.destroyers():
                context = ControllingExecutorContext(parent_context)
                context.entity = self.entity
                pe = ProcessExecutor(self.context, command)
                pe.execute(context)


class CollectingExecutor(Executor):

    def __init__(self, context: Context, entity: Entity, collectors: List[Collector], cgroups: List[ControllerGroup]):
        super().__init__(context, entity)
        self._collectors = collectors
        self._cgroups = cgroups
        self._executors = list()

    @property
    def collectors(self) -> List[Collector]:
        return self._collectors

    @collectors.setter
    def collectors(self, values: List[Collector]) -> None:
        self._collectors = values

    @property
    def cgroups(self) -> List[ControllerGroup]:
        return self._cgroups

    @cgroups.setter
    def cgroups(self, values: List[ControllerGroup]) -> None:
        self._cgroups = values

    @property
    def executors(self) -> List[Executor]:
        return self._executors

    def execute_main(self, parent_context: ExecutorContext) -> None:
        try:
            self._start_collectors(parent_context)
            for child in self.children:
                context = CollectingExecutorContext(parent_context)
                context.entity = self.entity
                child.execute(context)
        finally:
            self._stop_collectors()

    def _start_collectors(self, parent_context: ExecutorContext):
        for collector in self.collectors:
            context = CollectingExecutorContext(parent_context)
            context.entity = self.entity
            context.cgroups = self.cgroups
            collector.start(context)

    def _stop_collectors(self):
        for collector in self.collectors:
            collector.stop()


class ReplicatingExecutor(Executor):

    def execute_main(self, parent_context: ExecutorContext) -> None:
        for r in range(1, self.entity.config.replication + 1):
            self.context.logger.debug("{n} Execute[replication={r}]".format(n=self.entity, r=r))
            context = ReplicatingExecutorContext(parent_context)
            context.entity = self.entity
            context.current = r
            stat = self.stat()
            stat.start()
            try:
                self.execute_pre(context)
                for child in self.children:
                    child.execute(context)
            finally:
                # Try to perform post / clean up actions
                self.execute_post(context)
            stat.stop()
            stat.log(self.run_path(context))
            self.context.logger.debug("{n} Finish[replication={r} duration={d}]".format(n=self.entity, r=r, d=stat.timer.duration))

class ConcurrentExecutor(Executor):

    def execute_main(self, parent_context: ExecutorContext) -> None:
        if self.entity.config.concurrency:
            threads = list()
            self.context.logger.debug("{n} Execute[concurrency=true]".format(n=self.entity))
            c = 1
            for child in self.children:
                context = ConcurrentExecutorContext(parent_context)
                context.entity = self.entity
                context.current = c
                c += 1
                threads.append(threading.Thread(target=child.execute, args=[context]))
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
        else:
            self.context.logger.debug("{n} Execute[concurrency=false]".format(n=self.entity))
            for child in self.children:
                context = ConcurrentExecutorContext(parent_context)
                context.entity = self.entity
                context.current = 1
                child.execute(context)


class IteratingExecutor(Executor):

    def execute_main(self, parent_context: ExecutorContext) -> None:
        iter_vars = self._iter_vars()
        if len(iter_vars) > 0:
            # generate groups of iter values based on configuration
            iteration_type = self.entity.config.iteration
            if iteration_type == EntityConfig.ITERATION_TYPE_PRODUCT:
                iter_groups = itertools.product(*iter_vars.values())
            elif iteration_type == EntityConfig.ITERATION_TYPE_ZIP:
                iter_groups = zip(*iter_vars.values())
            else:
                raise Exception("Unknown iteration type {t}".format(t=iteration_type))
            # iterate with each group
            i = 1
            for iter_group in iter_groups:
                iters = dict(zip(iter_vars.keys(), iter_group))
                self.context.logger.debug("{n} Execute[iteration={i} iters={iters}]".format(n=self.entity, i=i, iters=iters))
                for child in self.children:
                    context = IteratingExecutorContext(parent_context)
                    context.entity = self.entity
                    context.current = i
                    context.iters = iters
                    child.execute(context)
                    i += 1
        else:
            self.context.logger.debug("{n} Execute[iteration=false]".format(n=self.entity))
            for child in self.children:
                context = IteratingExecutorContext(parent_context)
                context.entity = self.entity
                context.current = 1
                child.execute(context)

    def _iter_vars(self) -> Dict[str, List]:
        iter_vars = dict()
        e = self.entity
        while e is not None:
            if len(e.config.iters) > 0:
                for key, val in e.config.iters.items():
                    if key not in iter_vars:
                        iter_vars[key] = val
            e = e.parent
        return iter_vars


class ParallelExecutor(Executor):

    def execute_main(self, parent_context: ExecutorContext) -> None:
        parallelism = self.entity.config.parallelism
        if parallelism > 1:
            self.context.logger.debug("{n} Execute[parallelism={p}]".format(n=self.entity, p=parallelism))
            for child in self.children:
                threads = list()
                for p in range(1, parallelism + 1):
                    # If executor hierarchy is changed, deepcopy will be needed.
                    child_copy = copy.copy(child)
                    context = ParallelExecutorContext(parent_context)
                    context.entity = self.entity
                    context.current = p
                    threads.append(threading.Thread(target=child_copy.execute, args=[context]))
                for thread in threads:
                    thread.start()
                for thread in threads:
                    thread.join()
        else:
            self.context.logger.debug("{n} Execute[parallelism=false]".format(n=self.entity))
            for child in self.children:
                context = ParallelExecutorContext(parent_context)
                context.entity = self.entity
                context.current = 1
                child.execute(context)


class ProcessExecutor(Executor):

    def __init__(self, context: Context, entity: EntityCommand):
        super().__init__(context, entity)
        self._process = None

    def execute_main(self, parent_context: ExecutorContext) -> None:
        self._process = Process(self.context)
        self._process.command = self._command(parent_context)
        self._process.run()

    def _command(self, parent_context: ExecutorContext) -> EntityCommand:
        command = copy.copy(self.entity)
        command.cmd = self._prepare(command.cmd, parent_context)
        if command.become_cmd is not None:
            command.become_cmd = self._prepare_placeholders(command.become_cmd, parent_context)
        if command.pipe_stdout is not None:
            command.pipe_stdout = self._prepare_placeholders(command.pipe_stdout, parent_context)
        if command.pipe_stderr is not None:
            command.pipe_stderr = self._prepare_placeholders(command.pipe_stderr, parent_context)
        return command

    def _prepare(self, cmd, parent_context: ExecutorContext) -> str:
        cmd = self._prepare_placeholders(cmd, parent_context)
        cmd = self._sub_cgroup(cmd)
        cmd = self._sub_become(cmd)
        return cmd

    def _prepare_placeholders(self, cmd, parent_context: ExecutorContext) -> str:
        cmd = self._sub_context(cmd, parent_context)
        cmd = self._sub_entity(cmd)
        cmd = self._sub_params(cmd)
        cmd = self._sub_iters(cmd, parent_context)
        return cmd

    def _sub_context(self, cmd: str, parent_context: ExecutorContext) -> str:
        cmd = re.sub("{m:context:run_path}", self.run_path(parent_context), cmd)
        cmd = re.sub("{m:context:pid}", str(os.getpid()), cmd)
        cmd = re.sub("{m:context:ppid}", str(os.getppid()), cmd)
        cmd = re.sub("{m:context:pgid}", str(os.getpgid(os.getpid())), cmd)
        return cmd

    def _sub_entity(self, cmd: str) -> str:
        cmd = re.sub("{m:entity:command}", self.entity.name, cmd)
        # Support commands under pre and post sections
        if type(self.entity.parent) is EntityCase:
            cmd = re.sub("{m:entity:case}", self.entity.parent.name, cmd)
            cmd = re.sub("{m:entity:suite}", self.entity.parent.parent.name, cmd)
            cmd = re.sub("{m:entity:plan}", self.entity.parent.parent.parent.name, cmd)
        if type(self.entity.parent) is EntitySuite:
            cmd = re.sub("{m:entity:suite}", self.entity.parent.name, cmd)
            cmd = re.sub("{m:entity:plan}", self.entity.parent.parent.name, cmd)
        if type(self.entity.parent) is EntityPlan:
            cmd = re.sub("{m:entity:plan}", self.entity.parent.name, cmd)
        return cmd

    def _sub_params(self, cmd: str) -> str:
        e = self.entity
        while e is not None:
            for key, val in e.config.params.items():
                placeholder = "{{m:param:{key}}}".format(key=key)
                cmd = re.sub(re.escape(placeholder), val, cmd)
            e = e.parent
        return cmd

    def _sub_iters(self, cmd: str, parent_context: ExecutorContext) -> str:
        c = parent_context
        while c is not None:
            if hasattr(c, "iters") and c.iters is not None:
                for key, val in c.iters.items():
                    placeholder = "{{m:iter:{key}}}".format(key=key)
                    cmd = re.sub(re.escape(placeholder), val, cmd)
            c = c.parent
        return cmd

    def _sub_cgroup(self, cmd: str) -> str:
        cgroup_names = self.entity.cgroups
        if len(cgroup_names) > 0:
            parts = list()
            cgroups = self.entity.parent.parent.parent.cgroups
            for cgroup_name in cgroup_names:
                for cgroup in cgroups:
                    if cgroup.name == cgroup_name:
                        controllers = [c.name for c in cgroup.controllers]
                        parts.append("-g {controllers}:{cgroup_name}".format(controllers=",".join(controllers), cgroup_name=cgroup_name))
                        break
            if len(parts) == 0:
                raise Exception("No matching controller group found")
            cmd = "cgexec {parts} {cmd}".format(parts=" ".join(parts), cmd=cmd)
        return cmd

    def _sub_become(self, cmd: str) -> str:
        become_cmd = self.entity.become_cmd
        if become_cmd is not None:
            cmd = "{become_cmd} {cmd}".format(become_cmd=become_cmd, cmd=cmd)
        return cmd


class AsyncProcessExecutor(ProcessExecutor):

    def __init__(self, context: Context, entity: EntityCommand):
        super().__init__(context, entity)

    def execute_main(self, parent_context: ExecutorContext) -> None:
        self._process = Process(self.context)
        self._process.command = self._command(parent_context)
        self._process.start()

    def execute_stop(self) -> None:
        self._process.signal()
        self._process.wait()


class AsyncThreadExecutor(Executor):

    def __init__(self, context: Context, entity: Entity):
        super().__init__(context, entity)
        self._thread = None
        self._target = None

    @property
    def target(self) -> Any:
        return self._target

    @target.setter
    def target(self, value: Any) -> None:
        self._target = value

    def execute_main(self, parent_context: ExecutorContext):
        self._thread = Thread(self.context)
        self._thread.run(self._target, [parent_context])

    def execute_stop(self):
        self._target.join()
        self._thread.join()



from typing import List, Dict
from collections import defaultdict
from pymergen.collector.process import CollectorProcess


class CollectorPerf(CollectorProcess):

    def __init__(self):
        super().__init__()
        self._cmd_parts = list()
        self._custom = list()

    @property
    def cmd(self) -> str:
        if self._cmd is not None:
            return self._cmd
        self._prepare_cmd()
        self._cmd = " ".join(self._cmd_parts)
        return self._cmd

    def parse(self, config: Dict) -> None:
        super().parse(config)
        self._custom = config.get("custom", list())

    def _prepare_cmd(self):
        self._prepare_cmd_parts()

    def _prepare_cmd_parts(self):
        if self._custom is not None:
            self._cmd_parts.extend(self._custom)


class CollectorPerfEvent(CollectorPerf):

    def __init__(self):
        super().__init__()
        self._cgroup_events = defaultdict(list)
        self._system_events = list()

    @property
    def cgroup_events(self) -> Dict:
        return self._cgroup_events

    @cgroup_events.setter
    def cgroup_events(self, events: Dict) -> None:
        self._cgroup_events = events

    def add_cgroup_event(self, cgroup: str, name: str) -> None:
        self._cgroup_events[cgroup].append(name)

    @property
    def system_events(self) -> List:
        return self._system_events

    @system_events.setter
    def system_events(self, events: List) -> None:
        self._system_events = events

    def add_system_event(self, name: str) -> None:
        self._system_events.append(name)

    def parse(self, config: Dict) -> None:
        super().parse(config)
        for e in config["events"]:
            if "cgroup" in e:
                self.add_cgroup_event(e["cgroup"], e["name"])
            else:
                self.add_system_event(e["name"])

    def _prepare_cmd_parts(self):
        for cgroup, events in self.cgroup_events.items():
            self._cmd_parts.extend(["-e", "'{{{e}}}'".format(e=",".join(events)), "-G", cgroup])
        if len(self.system_events) > 0:
            self._cmd_parts.extend(["-a", "-e", "'{{{e}}}'".format(e=",".join(self.system_events))])
        super()._prepare_cmd_parts()


class CollectorPerfStat(CollectorPerfEvent):

    def __init__(self):
        super().__init__()

    def _prepare_cmd_parts(self):
        self._cmd_parts.extend(["perf", "stat", "record"])
        self._cmd_parts.extend(["-o", "{m:context:run_path}/collector.perf_stat.data"])
        super()._prepare_cmd_parts()


class CollectorPerfProfile(CollectorPerfEvent):

    def __init__(self):
        super().__init__()

    def _prepare_cmd_parts(self):
        self._cmd_parts.extend(["perf", "record"])
        self._cmd_parts.extend(["-o", "{m:context:run_path}/collector.perf_profile.data"])
        super()._prepare_cmd_parts()



from typing import List
from pymergen.entity.entity import Entity


class EntityCommand(Entity):

    def __init__(self):
        super().__init__()
        self._cmd = None
        self._become_cmd = None
        self._raise_error = True
        self._run_time = 0
        self._shell = False
        self._shell_executable = None
        self._timeout = None
        self._pipe_stdout = None
        self._pipe_stderr = None
        self._debug_stdout = False
        self._debug_stderr = False
        self._cgroups = list()

    @property
    def cmd(self) -> str:
        return self._cmd

    @cmd.setter
    def cmd(self, value: str) -> None:
        self._cmd = value

    @property
    def become_cmd(self) -> str:
        return self._become_cmd

    @become_cmd.setter
    def become_cmd(self, value: str) -> None:
        self._become_cmd = value

    @property
    def raise_error(self) -> bool:
        return self._raise_error

    @raise_error.setter
    def raise_error(self, value: bool) -> None:
        self._raise_error = value

    @property
    def run_time(self) -> int:
        return self._run_time

    @run_time.setter
    def run_time(self, value: int) -> None:
        self._run_time = value

    @property
    def shell(self) -> bool:
        return self._shell

    @shell.setter
    def shell(self, value: bool) -> None:
        self._shell = value

    @property
    def shell_executable(self) -> str:
        return self._shell_executable

    @shell_executable.setter
    def shell_executable(self, value: str) -> None:
        self._shell_executable = value

    @property
    def timeout(self) -> int:
        return self._timeout

    @timeout.setter
    def timeout(self, value: int) -> None:
        self._timeout = value

    @property
    def pipe_stdout(self) -> str:
        return self._pipe_stdout

    @pipe_stdout.setter
    def pipe_stdout(self, value: str) -> None:
        self._pipe_stdout = value

    @property
    def pipe_stderr(self) -> str:
        return self._pipe_stderr

    @pipe_stderr.setter
    def pipe_stderr(self, value: str) -> None:
        self._pipe_stderr = value

    @property
    def debug_stdout(self) -> bool:
        return self._debug_stdout

    @debug_stdout.setter
    def debug_stdout(self, value: bool) -> None:
        self._debug_stdout = value

    @property
    def debug_stderr(self) -> bool:
        return self._debug_stderr

    @debug_stderr.setter
    def debug_stderr(self, value: bool) -> None:
        self._debug_stderr = value

    @property
    def cgroups(self) -> List[str]:
        return self._cgroups

    @cgroups.setter
    def cgroups(self, values: List[str]) -> None:
        self._cgroups = values

    def dir_name(self) -> str:
        return "command_{command}".format(command=self.name)

    def short_name(self) -> str:
        return "Command[{command}]".format(command=self.name)

    def long_name(self) -> str:
        return "Plan[{plan}] > Suite[{suite}] > Case[{case}] > Command[{command}]".format(plan=self.parent.parent.parent.name, suite=self.parent.parent.name, case=self.parent.name, command=self.name)



from typing import List, Dict
from collections import defaultdict
from pymergen.plugin.plugin import Plugin


class PluginRegistry:

    def __init__(self):
        self._categories = list()
        self._plugins = defaultdict(dict)

    @property
    def categories(self) -> List:
        return self._categories

    @categories.setter
    def categories(self, values: List[str]) -> None:
        self._categories = values

    def add_category(self, value: str) -> None:
        self._categories.append(value)

    def add_plugin(self, category: str, engine: str, plugin: Plugin):
        self._check_category(category)
        self._plugins[category][engine] = plugin

    def get_plugin(self, category: str, engine: str) -> Plugin:
        self._check_category(category)
        return self._plugins.get(category).get(engine)

    def get_plugins(self, category: str) -> Dict[str, Plugin]:
        self._check_category(category)
        return self._plugins.get(category)

    def _check_category(self, category: str) -> None:
        if category not in self._categories:
            raise Exception("Plugin category {category} is not recognized".format(category=category))



from typing import Dict
from pymergen.collector.collector import Collector
from pymergen.core.executor import ExecutorContext, AsyncThreadExecutor


class CollectorThread(Collector):

    DEFAULT_RAMP = 0
    DEFAULT_INTERVAL = 1

    def __init__(self):
        super().__init__()
        self._executor = None
        self._join = False
        self._ramp = self.DEFAULT_RAMP
        self._interval = self.DEFAULT_INTERVAL

    def parse(self, config: Dict) -> None:
        super().parse(config)
        self._ramp = config.get("ramp", self.DEFAULT_RAMP)
        self._interval = config.get("interval", self.DEFAULT_INTERVAL)

    @property
    def ramp(self) -> int:
        return self._ramp

    @ramp.setter
    def ramp(self, value: int) -> None:
        self._ramp = value

    @property
    def interval(self) -> int:
        return self._interval

    @interval.setter
    def interval(self, value: int) -> None:
        self._interval = value

    def start(self, parent_context: ExecutorContext) -> None:
        self._executor = AsyncThreadExecutor(self.context, parent_context.entity)
        self._executor.target = self
        self._executor.execute(parent_context)

    def stop(self) -> None:
        self._executor.execute_stop()
        # Reset value for subsequent runs
        self._join = False

    def run(self, parent_context: ExecutorContext) -> None:
        raise NotImplementedError()

    def join(self) -> None:
        self._join = True
